/* Main source file for bprof. */

#include <unistd.h>		// getopt()
#include <string>
#include <map>
#include <set>
#include <algorithm>

using namespace std;

#include "bmonout.h"
#include "execute.h"
#include "sources.h"

/* Do I really have to write all this to use for_each() on all objects in a
   map?  If so, that shows you why functional languages are better than this
   feeble attempt of C++ at doing something similar. */

template <class T, class U>
class sec_mem_fun_t {
    void (T::*f)(U) const;
    U arg;
public:
    sec_mem_fun_t(void (T::*m)(U) const, U& x) : f(m), arg(x) {}
    template <class F>
    void operator() (const pair<F, T*>& x) { (x.second->*f)(arg); }
};

template <class T, class U>
inline sec_mem_fun_t<T, U> sec_mem_fun(void (T::*m)(U) const, U& x)
{
    return sec_mem_fun_t<T, U>(m, x);
}

int main(int argc, char* argv[])
{
    string sffx(".bprof");	// Default suffix
    set<dir> gooddirs;
    dir pwd(".");
    if (pwd.valid()) {
	gooddirs.insert(pwd);	// Always use pwd
    }
    int verbose = 0;		// Whether to print filenames
    unsigned long int totalcount = 0;

    int option;
    while ((option = getopt(argc, argv, "d:s:v")) != EOF) {
	switch (option) {
	case 'd': 
	{
	    dir extradir(optarg);
	    if (extradir.valid()) {
		gooddirs.insert(extradir);
	    } else {
		cerr << "Cannot find directory " << optarg << ", proceeding without\n";
	    }
	    break;
	}
	case 's':
	    sffx = optarg;
	    break;
	case 'v':
	    verbose = sourcefile::verboseflag;
	    break;
	case '?':
	    exit(1);
	}
    }

    // First non-option argument is executable
    const char* filename;
    if (optind < argc) {
	filename = argv[optind++];
    } else {
	filename = "a.out";
    }
    executable aout(filename);
    
    // Other non-options args are bmon.out files
    int numbmon = argc - optind;
    bmonout** bmon;		// Array of bmonout pointers
    if (numbmon) {
	bmon = new bmonout*[numbmon];
	for (int i = 0; i < numbmon; i++) {
	    bmon[i] = new bmonout(argv[optind]);
	    if (bmon[i]->mtime() < aout.mtime()) {
		cerr << "Warning: executable is newer than " << argv[optind] << '\n';
	    }
	    optind++;
	}
    } else {
	// Perhaps we should read all bmon.out.<num> if none are specified.
	numbmon = 1;
	bmon = new bmonout*[numbmon];
	bmon[0] = new bmonout("bmon.out.0");
	if (bmon[0]->mtime() < aout.mtime()) {
	    cerr << "Warning: executable is newer than bmon.out.0\n";
	}
    }

    typedef map<string, sfpnt> sourcemap;
    sourcemap source;

    // Inner loop should be over bmon files to speed things up,
    // although I tend to use only a single bmon file anyway.
    for (int i = 0; i < numbmon; i++) {
	totalcount += bmon[i]->totalcount();
	int inum = bmon[i]->countnum();
	for (int j = 0; j < inum; j++) {
	    caddr_t eip = (*bmon[i])[j].pc;
	    const char* thisname;
	    unsigned int lineno;
	    if (!aout.find_line(eip, &thisname, &lineno)) {
		cerr << "Could not find line for eip\n";
		continue;
	    }
	    if (!thisname) continue;
	    string cname(thisname);
	    // If filename unknown, add to source set.
	    if (!source.count(cname)) {
		int lastslash = cname.rfind("/");
		string dirname = cname.substr(0, lastslash); // Final -1 ???
		dir filedir(dirname.c_str());
		int really = gooddirs.count(filedir) ? sourcefile::reallyflag : 0;
		sourcemap::value_type newkey(cname, new sourcefile(cname, really | verbose));
		source.insert(newkey);
		if (aout.mtime() < source[cname]->mtime())
		    cerr << "Warning: source file " << cname << " newer than executable\n";
	    }
	    sourcefile& sfile = *source[cname];
	    sfile[lineno] += (*bmon[i])[j].count;
	}
    }

    for (int i = 0; i < numbmon; i++) {
	delete bmon[i];
    }
    delete[] bmon;

    if (verbose) {
	cout << "Total number of counts is " << totalcount << '\n';
    }

    // Write the source files with data
#if 1
    // The for_each version using the sec_mem_fun_t defined above.
    for_each(source.begin(), source.end(),
             sec_mem_fun(&sourcefile::paste, sffx));
#else
    // The old-fashioned version using for.
    for (sourcemap::const_iterator p = source.begin(); p != source.end(); p++) {
	p->second->paste(sffx);
    }
#endif

    return 0;
}
